import numpy as np
from numba import jit

def compute_energy(sigmas, y, lam, rho):
    E = -np.log((sigmas * y).sum(axis=1)).mean()
    E -= lam * (rho * np.log(y.mean(axis=0))).sum()
    
    return E 

def compute_soft_cardinality(Y):
    N, K = Y.shape
    eps = 1e-50
    denominator = np.sum(Y, axis=0) + eps
    soft_cardinality = Y / denominator.reshape((1,K))
    
    return soft_cardinality

@jit(nopython=True) 
def fx(x, sigmas, S, lam, rho):
    eps = 1e-50
    rhoS = rho * S
    return (lam * rhoS / (lam * rhoS.sum() + 1 - sigmas / (x + eps))).sum() - 1

@jit(nopython=True)
def jx(x, sigmas, S, lam, rho):
    rhoS = rho * S
    return (-lam * rhoS * sigmas / (lam * rhoS.sum() * x + x - sigmas)**2).sum()
    
@jit(nopython=True)
def compute_pseudo_confidence(sigmas, S, lam, rho):
    tau = 1e-10
    eps = 1e-50
    N, K = sigmas.shape
    Y = np.zeros_like(sigmas)
    
    for i in range(N):
        sigma = sigmas[i]
        Si = S[i]
        rhoSi = rho * Si
        
        x = (sigma / (lam * rhoSi.sum()-lam * rhoSi + 1.0)).max()
        fvalue = fx(x, sigma, Si, lam, rho)
        while np.abs(fvalue) > tau:
            fvalue = fx(x, sigma, Si, lam, rho)
            dvalue = jx(x, sigma, Si, lam, rho)
            x -= fvalue/(dvalue + eps)
        Y[i,:] = lam * rhoSi / (lam * rhoSi.sum() + 1.0 - sigma / (x + eps))
        
    return Y

def y_CCE_Update(sigmas, lam, rho):
    tau = 1e-2
    N, K = sigmas.shape
    Y_prev = np.ones_like(sigmas)
    Y = sigmas # initialization
    Es = []
    E_prev = np.infty
    E = compute_energy(sigmas, Y, lam, rho)
    Es.append(E.item())
    itertime = []
    
    steps = 0
    while np.abs(E - E_prev)>=1e-4: #LA.norm(Y-Y_prev,ord='fro')>=tau:# and steps <=1:
        s = time.time()
        S = compute_soft_cardinality(Y)
        Y_prev = Y
        Y = compute_pseudo_confidence(sigmas, S, lam * N, rho)
        E_prev  =E
        E = compute_energy(sigmas, Y, lam, rho)
        Es.append(E.item())
        steps += 1
        e = time.time()
        itertime.append(e-s)
        
    return Y, Es, np.array(itertime).mean()